r"""This module implements an asset manager."""

from __future__ import annotations

__all__ = ["AssetExistsError", "AssetManager", "AssetNotFoundError"]

import copy
import logging
from typing import Any

from coola import objects_are_equal, summary
from coola.utils import str_indent, str_mapping

logger = logging.getLogger(__name__)


class AssetExistsError(Exception):
    r"""Raised when trying to add an asset that already exists."""


class AssetNotFoundError(Exception):
    r"""Raised when trying to access an asset that does not exist."""


class AssetManager:
    r"""Implements an asset manager.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.asset import AssetManager
        >>> manager = AssetManager()
        >>> manager
        AssetManager()
        >>> manager.add_asset("mean", 5)
        >>> manager
        AssetManager(
          (mean): <class 'int'> 5
        )
        >>> manager.get_asset("mean")
        5
    """

    def __init__(self, assets: dict[str, Any] | None = None) -> None:
        self._assets = assets or {}

    def __len__(self) -> int:
        return len(self._assets)

    def __repr__(self) -> str:
        assets = {name: summary(asset) for name, asset in self._assets.items()}
        args = f"\n  {str_indent(str_mapping(assets))}\n" if assets else ""
        return f"{self.__class__.__qualname__}({args})"

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(num_assets={len(self._assets):,})"

    def add_asset(self, name: str, asset: Any, replace_ok: bool = False) -> None:
        r"""Adds an asset to the asset manager.

        Note that the name should be unique. If the name exists, the
        old asset will be overwritten by the new asset.

        Args:
        ----
            name (str): Specifies the name of the asset to add.
            asset: Specifies the asset to add.
            replace_ok (bool, optional): If ``False``,
                ``AssetExistsError`` is raised if an asset with the
                same name exists. Default: ``False``

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
        """
        if name in self._assets and not replace_ok:
            raise AssetExistsError(
                f"`{name}` is already used to register an asset. "
                "Use `replace_ok=True` to replace an asset"
            )
        self._assets[name] = asset

    def clone(self) -> AssetManager:
        r"""Creates a deep copy of the current asset manager.

        Returns
        -------
            ``AssetManager``: A deep copy of the current asset manager.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager({"name": 5})
            >>> clone = manager.clone()
            >>> manager.add_asset("name", 7, replace_ok=True)
            >>> manager
            AssetManager(
              (name): <class 'int'>  7
            )
            >>> clone
            AssetManager(
              (name): <class 'int'>  5
            )
        """
        return AssetManager(copy.deepcopy(self._assets))

    def equal(self, other: Any) -> bool:
        r"""Indicates if two objects are equal.

        Args:
        ----
            other: Specifies the object to compare with.

        Returns:
        -------
            bool: ``True`` if the two objects are equal, otherwise
                ``False``.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
            >>> manager2 = AssetManager()
            >>> manager.equal(manager2)
            False
            >>> manager2.add_asset("mean", 5)
            >>> manager.equal(manager2)
            True
        """
        if not isinstance(other, AssetManager):
            return False
        return objects_are_equal(self._assets, other._assets)

    def get_asset(self, name: str) -> Any:
        r"""Gets an asset.

        Args:
        ----
            name (str): Specifies the asset to get.

        Returns:
        -------
            The asset

        Raises:
        ------
            ``AssetNotFoundError`` if the asset does not exist.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
            >>> manager.get_asset("mean")
            5
        """
        if name not in self._assets:
            raise AssetNotFoundError(f"The asset '{name}' does not exist")
        return self._assets[name]

    def get_asset_names(self) -> tuple[str, ...]:
        r"""Gets all the asset names.

        Returns
        -------
            tuple: The asset names.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
            >>> manager.get_asset_names()
            ('mean',)
        """
        return tuple(self._assets.keys())

    def has_asset(self, name: str) -> bool:
        r"""Indicates if the asset exists or not.

        Args:
        ----
            name (str): Specifies the name of the asset.

        Returns:
        -------
            bool: ``True`` if the asset exists, otherwise ``False``

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.has_asset("mean")
            False
            >>> manager.add_asset("mean", 5)
            >>> manager.has_asset("mean")
            True
        """
        return name in self._assets

    def remove_asset(self, name: str) -> None:
        r"""Removes an asset.

        Args:
        ----
            name (str): Specifies the name of the asset to remove.

        Raises:
        ------
            ``AssetNotFoundError`` if the asset does not exist.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
            >>> manager.remove_asset("mean")
            >>> manager.has_asset("mean")
            False
        """
        if name not in self._assets:
            raise AssetNotFoundError(
                f"The asset '{name}' does not exist so it is not possible to remove it"
            )
        del self._assets[name]

    def load_state_dict(self, state_dict: dict, keys: list | tuple | None = None) -> None:
        r"""Loads the state dict of each module.

        Note this method ignore the missing modules or keys. For
        example if you want to load the optimizer module but there is
        no 'optimizer' key in the state dict, this method will ignore
        the optimizer module.

        Args:
        ----
            state_dict (dict): Specifies the state dict to load.
            keys (list or tuple or ``None``): Specifies the keys to
                load. If ``None``, it loads all the keys associated
                to the registered modules.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from torch import nn
            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("my_module", nn.Linear(4, 6))
            >>> manager.load_state_dict(
            ...     {"my_module": {"weight": torch.ones(6, 4), "bias": torch.zeros(6)}}
            ... )
        """
        keys = keys or tuple(self._assets.keys())
        for key in keys:
            if key not in state_dict:
                logger.info(f"Ignore key {key} because it is not in the state dict")
                continue
            if key not in self._assets:
                logger.info(f"Ignore key {key} because there is no module associated to it")
                continue
            if not hasattr(self._assets[key], "load_state_dict"):
                logger.info(
                    f"Ignore key {key} because the module does not have 'load_state_dict' method"
                )
                continue
            self._assets[key].load_state_dict(state_dict[key])

    def state_dict(self) -> dict:
        r"""Creates a state dict with all the modules.

        The state of each module is store with the associated key of
        the module.

        Returns
        -------
            dict: The state dict of all the modules.

        Example usage:

        .. code-block:: pycon

            >>> from torch import nn
            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("my_module", nn.Linear(4, 6))
            >>> manager.state_dict()
            {'my_module': OrderedDict([('weight', tensor([[...)), ('bias', tensor([...))])}
            >>> manager.add_asset("int", 123)
            >>> manager.state_dict()
            {'my_module': OrderedDict([('weight', tensor([[...]])), ('bias', tensor([...]))])}
        """
        state = {}
        for name, module in self._assets.items():
            if hasattr(module, "state_dict"):
                state[name] = module.state_dict()
            else:
                logger.info(f"Skip '{name}' module because it does not have 'state_dict' method")
        return state



from __future__ import annotations

__all__ = ["BaseDataStreamCreator", "is_datastream_creator_config", "setup_datastream_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.datastreams import BaseDataStream
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDataStreamCreator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a datastream creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datastream import IterableDataStreamCreator
        >>> creator = IterableDataStreamCreator((1, 2, 3, 4, 5))
        >>> creator
        IterableDataStreamCreator(cache=False, length=5)
        >>> datastream = creator.create()
        >>> datastream
        IterableDataStream(length=5)
    """

    @abstractmethod
    def create(self, engine: BaseEngine | None = None) -> BaseDataStream[T]:
        r"""Create a datastreams.

        Args:
        ----
            engine (``gravitorch.engines.BaseEngine`` or ``None``,
                optional): Specifies an engine. Default: ``None``

        Returns:
        -------
            ``BaseDataStream``: The created datastreams.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.creators.datastream import IterableDataStreamCreator
            >>> creator = IterableDataStreamCreator((1, 2, 3, 4, 5))
            >>> datastream = creator.create()
            >>> datastream
            IterableDataStream(length=5)
        """


def is_datastream_creator_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseDataStreamCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseDataStreamCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datastream import is_datastream_creator_config
        >>> is_datastream_creator_config(
        ...     {"_target_": "gravitorch.creators.datastream.IterableDataStreamCreator"}
        ... )
        True
    """
    return is_object_config(config, BaseDataStreamCreator)


def setup_datastream_creator(creator: BaseDataStreamCreator[T] | dict) -> BaseDataStreamCreator[T]:
    r"""Sets up the datastream creator.

    The datastream creator is instantiated from its configuration by
    using the ``BaseDataStreamCreator`` factory function.

    Args:
    ----
        creator (``BaseDataStreamCreator`` or dict): Specifies the
            datastream creator or its configuration.

    Returns:
    -------
        ``BaseDataStreamCreator``: The instantiated datastream creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datastream import setup_datastream_creator
        >>> creator = setup_datastream_creator(
        ...     {
        ...         "_target_": "gravitorch.creators.datastream.IterableDataStreamCreator",
        ...         "iterable": (1, 2, 3, 4, 5),
        ...     }
        ... )
        >>> creator
        IterableDataStreamCreator(cache=False, length=5)
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing the datastream creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDataStreamCreator.factory(**creator)
    return creator



from __future__ import annotations

__all__ = [
    "BaseDataLoader2Creator",
    "DataLoader2Creator",
    "VanillaDataLoader2Creator",
    "is_dataloader2_creator_config",
    "setup_dataloader2_creator",
]

from gravitorch.creators.dataloader2.base import (
    BaseDataLoader2Creator,
    is_dataloader2_creator_config,
    setup_dataloader2_creator,
)
from gravitorch.creators.dataloader2.vanilla import (
    DataLoader2Creator,
    VanillaDataLoader2Creator,
)



from __future__ import annotations

__all__ = ["DataLoaderDataStreamCreator"]

from typing import TYPE_CHECKING, TypeVar

from coola.utils import str_indent, str_mapping
from torch.utils.data import DataLoader

from gravitorch.creators.datastream.base import BaseDataStreamCreator
from gravitorch.dataloaders.factory import is_dataloader_config
from gravitorch.datastreams.dataloader import DataLoaderDataStream
from gravitorch.experimental.dataloader.base import (
    BaseDataLoaderCreator,
    setup_dataloader_creator,
)
from gravitorch.experimental.dataloader.vanilla import DataLoaderCreator

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class DataLoaderDataStreamCreator(BaseDataStreamCreator[T]):
    r"""Implements a simple ``DataLoaderDataStream`` creator.

    Args:
    ----
        dataloader (``torch.utils.data.DataLoader`` or
            ``BaseDataLoaderCreator``): Specifies a dataloader (or its
            configuration) or a dataloader creator (or its
            configuration).

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datasets import ExampleDataset
        >>> from gravitorch.creators.datastream import DataLoaderDataStreamCreator
        >>> from torch.utils.data import DataLoader
        >>> creator = DataLoaderDataStreamCreator(DataLoader(ExampleDataset([1, 2, 3, 4, 5])))
        >>> creator
        DataLoaderDataStreamCreator(
          (dataloader): DataLoaderCreator(
              cache=False
              dataloader=<torch.utils.data.dataloader.DataLoader object at 0x...>
            )
        )
        >>> datastream = creator.create()
        >>> datastream
        DataLoaderDataStream(length=5)
    """

    def __init__(self, dataloader: DataLoader | BaseDataLoaderCreator | dict) -> None:
        if isinstance(dataloader, DataLoader) or (
            isinstance(dataloader, dict) and is_dataloader_config(dataloader)
        ):
            dataloader = DataLoaderCreator(dataloader)
        self._dataloader = setup_dataloader_creator(dataloader)

    def __repr__(self) -> str:
        config = {"dataloader": self._dataloader}
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(config))}\n)"

    def create(self, engine: BaseEngine | None = None) -> DataLoaderDataStream[T]:
        return DataLoaderDataStream(self._dataloader.create(engine))



from __future__ import annotations

__all__ = ["IterableDataStreamCreator"]

from collections.abc import Iterable
from contextlib import suppress
from typing import TYPE_CHECKING, TypeVar

from gravitorch.creators.datastream.base import BaseDataStreamCreator
from gravitorch.datastreams.iterable import IterableDataStream
from gravitorch.utils.factory import setup_object
from gravitorch.utils.format import str_mapping

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class IterableDataStreamCreator(BaseDataStreamCreator[T]):
    r"""Implements a simple ``IterableDataStream`` creator.

    Args:
    ----
        iterable (``Iterable`` or dict): Specifies an iterable or its
            configuration.
        cache (bool, optional): If ``True``, the iterable is created
            only the first time, and then a copy of the iterable is
            returned for each call to the ``create`` method.
            Default: ``False``
        **kwargs: See ``IterableDataStream`` documentation.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datastream import IterableDataStreamCreator
        >>> creator = IterableDataStreamCreator((1, 2, 3, 4, 5))
        >>> creator
        IterableDataStreamCreator(cache=False, length=5)
        >>> datastream = creator.create()
        >>> datastream
        IterableDataStream(length=5)
    """

    def __init__(self, iterable: Iterable[T], cache: bool = False, **kwargs) -> None:
        self._iterable = iterable
        self._cache = bool(cache)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        config = {"cache": self._cache} | self._kwargs
        with suppress(TypeError):
            config["length"] = f"{len(self._iterable):,}"
        return (
            f"{self.__class__.__qualname__}({str_mapping(config, sorted_keys=True, one_line=True)})"
        )

    def create(self, engine: BaseEngine | None = None) -> IterableDataStream[T]:
        iterable = setup_object(self._iterable)
        if self._cache:
            self._iterable = iterable
        return IterableDataStream(iterable, **self._kwargs)



__all__ = [
    "BaseLRSchedulerCreator",
    "LRSchedulerCreator",
    "is_lr_scheduler_creator_config",
    "setup_lr_scheduler_creator",
]

from gravitorch.creators.lr_scheduler.base import BaseLRSchedulerCreator
from gravitorch.creators.lr_scheduler.factory import (
    is_lr_scheduler_creator_config,
    setup_lr_scheduler_creator,
)
from gravitorch.creators.lr_scheduler.vanilla import LRSchedulerCreator



from __future__ import annotations

__all__ = ["EpochRandomDataPipeCreator"]

import logging
from collections.abc import Sequence
from typing import TYPE_CHECKING, TypeVar

from coola.utils import str_indent, str_mapping
from objectory import OBJECT_TARGET, factory
from torch.utils.data import IterDataPipe, MapDataPipe

from gravitorch.creators.datapipe.base import BaseDataPipeCreator
from gravitorch.distributed import comm as dist
from gravitorch.utils.format import str_pretty_json

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class EpochRandomDataPipeCreator(BaseDataPipeCreator[T]):
    r"""Implements a ``DataPipe`` creator to create a ``DataPipe`` object
    where its random seed is controlled by an engine.

    Given an engine, the random seed is set based on the engine random
    seed, the current epoch value, the maximum number of epochs and
    the distributed rank.

    Args:
    ----
        config (dict): Specifies the configuration of the
            ``DataPipe``.
        random_seed_key (str, optional): Specifies the key in the
            configuration which is used to indicate the random seed
            of the ``DataPipe``. If this key exists in ``config``,
            it will be replaced by a new value, based on the engine
            state.

    Example usage:

    .. code-block:: pycon

        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.creators.datapipe import EpochRandomDataPipeCreator
        >>> from gravitorch.testing import create_dummy_engine
        >>> creator = EpochRandomDataPipeCreator(
        ...     {
        ...         "_target_": "gravitorch.datapipes.iter.TensorDictShuffler",
        ...         "datapipe": {
        ...             "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...             "iterable": [1, 2, 3, 4, 5],
        ...         },
        ...         "random_seed": 42,
        ...     }
        ... )
        >>> creator
        EpochRandomDataPipeCreator(
          (config): {
              "_target_": "gravitorch.datapipes.iter.TensorDictShuffler",
              "datapipe": {
                "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
                "iterable": [
                  1,
                  2,
                  3,
                  4,
                  5
                ]
              },
              "random_seed": 42
            }
          (random_seed_key): random_seed
        )
        >>> engine = create_dummy_engine()
        >>> dp = creator.create(engine)
        >>> dp
        TensorDictShufflerIterDataPipe(
          (dim): 0
          (random_seed): 41
          (datapipe): {'_target_': 'torch.utils.data.datapipes.iter.IterableWrapper', 'iterable': [1, 2, 3, 4, 5]}
        )
    """

    def __init__(self, config: dict, random_seed_key: str = "random_seed") -> None:
        self._config = config
        self._random_seed_key = str(random_seed_key)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "config": str_pretty_json(self._config),
                    "random_seed_key": self._random_seed_key,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def create(
        self, engine: BaseEngine | None = None, source_inputs: Sequence | None = None
    ) -> IterDataPipe[T] | MapDataPipe[T]:
        if engine is None:
            raise RuntimeError(
                "engine cannot be None because the epoch value is used to create the "
                "DataPipe object"
            )
        source_inputs = source_inputs or ()
        config = self._config.copy()  # Make a copy because the dict is modified below.
        target = config.pop(OBJECT_TARGET)
        config[self._random_seed_key] = (
            config.get(self._random_seed_key, engine.random_seed)
            + engine.epoch
            + engine.max_epochs * dist.get_rank()
        )
        logger.info(
            f"Set the random seed to {config[self._random_seed_key]}  [{OBJECT_TARGET}: {target}]"
        )
        return factory(target, *source_inputs, **config)



from __future__ import annotations

__all__ = ["NoOpProfiler"]

import logging
from types import TracebackType

from gravitorch.utils.profilers.base import BaseProfiler

logger = logging.getLogger(__name__)


class NoOpProfiler(BaseProfiler):
    r"""Implements a no-op profiler.

    This class allows to use a profiler without changing the
    implementation.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.utils.profilers import NoOpProfiler
        >>> with NoOpProfiler() as profiler:
        ...     x = torch.ones(2, 3)
        ...     for _ in range(20):
        ...         x += x
        ...         profiler.step()
        ...
    """

    def __enter__(self) -> NoOpProfiler:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def step(self) -> None:
        r"""Signals the profiler that the next profiling step has
        started.

        Because it is a no-op class, this method does nothing.
        """



from __future__ import annotations

__all__ = [
    "BaseDataStreamCreator",
    "IterableDataStreamCreator",
    "DataLoaderDataStreamCreator",
    "is_datastream_creator_config",
    "setup_datastream_creator",
]

from gravitorch.creators.datastream.base import (
    BaseDataStreamCreator,
    is_datastream_creator_config,
    setup_datastream_creator,
)
from gravitorch.creators.datastream.dataloader import DataLoaderDataStreamCreator
from gravitorch.creators.datastream.iterable import IterableDataStreamCreator



from __future__ import annotations

__all__ = ["CoreCreator"]

from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping
from torch import nn
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.creators.core.base import BaseCoreCreator

if TYPE_CHECKING:
    from gravitorch.datasources import BaseDataSource
    from gravitorch.engines import BaseEngine
    from gravitorch.lr_schedulers import LRSchedulerType


class CoreCreator(BaseCoreCreator):
    r"""Implements a simple core engine moules creator.

    This creator does not always "create" the core modules because
    they can already exist. The user is responsible to attach the
    core modules to the engine. This creator only adds the given
    modules to the engine state.

    Args:
    ----
        datasource (``BaseDataSource`` or dict): Specifies the data
            source or its configuration.
        model (``BaseModelCreator`` or dict): Specifies the model
            or its configuration.
        optimizer (``BaseOptimizerCreator`` or dict or ``None`):
            Specifies the optimizer or its configuration.
            Default: ``None``
        lr_scheduler (``BaseLRSchedulerCreator`` or dict or ``None`):
            Specifies the LR scheduler or its configuration.
            Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine
        >>> from gravitorch.creators.core import CoreCreator
        >>> creator = CoreCreator(
        ...     datasource={"_target_": "gravitorch.testing.DummyDataSource"},
        ...     model={"_target_": "gravitorch.testing.DummyClassificationModel"},
        ...     optimizer={"_target_": "torch.optim.SGD", "lr": 0.01},
        ...     lr_scheduler={"_target_": "torch.optim.lr_scheduler.StepLR", "step_size": 5},
        ... )
        >>> creator
        CoreCreator(
          (datasource): DummyDataSource(
              (datasets):
                (train): DummyDataset(num_examples=4, feature_size=4)
                (eval): DummyDataset(num_examples=4, feature_size=4)
              (dataloader_creators):
                (train): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
                (eval): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
            )
          (model): DummyClassificationModel(
              (linear): Linear(in_features=4, out_features=3, bias=True)
              (criterion): CrossEntropyLoss()
            )
          (optimizer): SGD (
            Parameter Group 0...
                lr: 0.01
                maximize: False
                momentum: 0
                nesterov: False
                weight_decay: 0
            )
          (lr_scheduler): <torch.optim.lr_scheduler.StepLR object at 0x...>
        )
        >>> engine = create_dummy_engine()
        >>> datasource, model, optimizer, lr_scheduler = creator.create(engine)
        >>> datasource
        DummyDataSource(
          (datasets):
            (train): DummyDataset(num_examples=4, feature_size=4)
            (eval): DummyDataset(num_examples=4, feature_size=4)
          (dataloader_creators):
            (train): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
            (eval): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
        )
        >>> model
        DummyClassificationModel(
          (linear): Linear(in_features=4, out_features=3, bias=True)
          (criterion): CrossEntropyLoss()
        )
        >>> optimizer
        SGD (
          Parameter Group 0...
              lr: 0.01
              maximize: False
              momentum: 0
              nesterov: False
              weight_decay: 0
        )
        >>> lr_scheduler
        <torch.optim.lr_scheduler.StepLR object at 0x...>
    """

    def __init__(
        self,
        datasource: BaseDataSource | dict,
        model: nn.Module | dict,
        optimizer: Optimizer | dict | None = None,
        lr_scheduler: LRSchedulerType | dict | None = None,
    ) -> None:
        # Local imports to avoid cyclic imports
        from gravitorch.datasources import setup_datasource
        from gravitorch.lr_schedulers import setup_lr_scheduler
        from gravitorch.models import setup_model
        from gravitorch.optimizers import setup_optimizer

        self._datasource = setup_datasource(datasource)
        self._model = setup_model(model)
        self._optimizer = setup_optimizer(model=self._model, optimizer=optimizer)
        self._lr_scheduler = setup_lr_scheduler(
            optimizer=self._optimizer, lr_scheduler=lr_scheduler
        )

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "datasource": self._datasource,
                    "model": self._model,
                    "optimizer": self._optimizer,
                    "lr_scheduler": self._lr_scheduler,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def create(
        self, engine: BaseEngine
    ) -> tuple[BaseDataSource, nn.Module, Optimizer | None, LRSchedulerType | None]:
        engine.add_module(ct.DATA_SOURCE, self._datasource)
        engine.add_module(ct.MODEL, self._model)
        if self._optimizer:
            engine.add_module(ct.OPTIMIZER, self._optimizer)
        if self._lr_scheduler:
            engine.add_module(ct.LR_SCHEDULER, self._lr_scheduler)
        return self._datasource, self._model, self._optimizer, self._lr_scheduler



from __future__ import annotations

__all__ = ["BaseDataLoaderCreator", "is_dataloader_creator_config", "setup_dataloader_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config
from torch.utils.data import DataLoader

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDataLoaderCreator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a dataloader creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataloader import VanillaDataLoaderCreator
        >>> creator = VanillaDataLoaderCreator(
        ...     {
        ...         "_target_": "gravitorch.datasets.DummyMultiClassDataset",
        ...         "num_examples": 10,
        ...         "num_classes": 2,
        ...         "feature_size": 4,
        ...     }
        ... )
        >>> creator.create()
        <torch.utils.data.dataloader.DataLoader object at 0x...>
    """

    @abstractmethod
    def create(self, engine: BaseEngine | None = None) -> DataLoader[T]:
        r"""Create a dataloader.

        Args:
        ----
            engine (``gravitorch.engines.BaseEngine`` or ``None``,
                optional): Specifies an engine. Default: ``None``

        Returns:
        -------
            ``torch.utils.data.DataLoader``: The created dataloader.
        """


def is_dataloader_creator_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseDataLoaderCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseDataLoaderCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataloader import is_dataloader_creator_config
        >>> is_dataloader_creator_config(
        ...     {"_target_": "gravitorch.experimental.dataloader.DataLoaderCreator"}
        ... )
        True
    """
    return is_object_config(config, BaseDataLoaderCreator)


def setup_dataloader_creator(creator: BaseDataLoaderCreator[T] | dict) -> BaseDataLoaderCreator[T]:
    r"""Sets up the dataloader creator.

    The dataloader creator is instantiated from its configuration by
    using the ``BaseDataLoaderCreator`` factory function.

    Args:
    ----
        creator (``BaseDataLoaderCreator`` or dict): Specifies the
            dataloader creator or its configuration.

    Returns:
    -------
        ``BaseDataLoaderCreator``: The instantiated dataloader creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataloader import setup_dataloader_creator
        >>> creator = setup_dataloader_creator(
        ...     {
        ...         "_target_": "gravitorch.experimental.dataloader.VanillaDataLoaderCreator",
        ...         "dataset": {
        ...             "_target_": "gravitorch.datasets.DummyMultiClassDataset",
        ...             "num_examples": 10,
        ...             "num_classes": 2,
        ...             "feature_size": 4,
        ...         },
        ...     }
        ... )
        >>> creator
        VanillaDataLoaderCreator(
          dataset=DatasetCreator(
              cache=False
              dataset={'_target_': 'gravitorch.datasets.DummyMultiClassDataset', 'num_examples': 10, 'num_classes': 2, 'feature_size': 4}
            )
          seed=0
        )
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing the dataloader creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDataLoaderCreator.factory(**creator)
    return creator



__all__ = [
    "BaseDataSourceCreator",
    "DataSourceCreator",
    "is_datasource_creator_config",
    "setup_datasource_creator",
]

from gravitorch.creators.datasource.base import (
    BaseDataSourceCreator,
    is_datasource_creator_config,
    setup_datasource_creator,
)
from gravitorch.creators.datasource.vanilla import DataSourceCreator
